/**
 * scan_xd87-4.ino
 * USBCore.* は、Arduino標準のものを改造して使う。
 * マトリックススキャンしてオン／オフされたキーのスキャンコードをシリアル出力する。
 * シリアルが使えるようになるまで、setup()で待つ。
 * 開始したら、RGB LEDを点灯し、キーオンオフがあったらしばらく明るくする。
 * BGLEDとCAPSLEDを別のタイミングで点滅させる。
 */
extern "C" {
#include "light_ws2812.h"
}

static const uint8_t ROW_COUNT = 6;
static const uint8_t COL_COUNT = 17; // ROWあたりのビット数(キー数）

// 1ROW あたり 17ビット
uint8_t key_pressed[ROW_COUNT * 3]; // 現在押下情報
uint8_t last_key_pressed[ROW_COUNT * 3];  // 前回押下情報

static const int16_t RGB_COUNT = 14;
static const int16_t LED_BUFFER_SIZE = RGB_COUNT * sizeof(cRGB);
uint8_t led_buffer[LED_BUFFER_SIZE];

static const uint8_t VAL_DIMMED = 0x10; // 暗い
static const uint8_t VAL_BRIGHT = 0x40; // 明るい

void rgb_on( uint8_t val) {
  memset(led_buffer,  val, sizeof(led_buffer));
  ws2812_sendarray_mask(led_buffer, LED_BUFFER_SIZE, _BV(PINB7));
}

// xd87
#define BGLED_ON()    PORTD &= ~_BV(0)
#define BGLED_OFF()   PORTD |= _BV(0)
#define CAPSLED_ON()    PORTE &= ~_BV(2)
#define CAPSLED_OFF()   PORTE |= _BV(2)

void setup() {
  cli();
  
  CLKPR = 0x80;
  CLKPR = 0;    // CLKPR = 1/1
  uint8_t v = MCUCR | 0x80;
  MCUCR = v;    // disable JTAG for PF
  MCUCR = v;
  
  TCCR1A = 0; // init timers (without TC0)
  TCCR3A = 0;
  TCCR4A = 0;

// setup Ports.
  DDRB =  _BV(7) | _BV(1) | _BV(0); // PB7 = RGBLED.
  DDRC =  _BV(7);
  DDRD =  _BV(5) | _BV(3) | _BV(1) | _BV(0);
  DDRE =  _BV(2);
  DDRF =  0;
// set INPUT_PULLUP to COLs.
  PORTB =  _BV(6) | _BV(5) | _BV(4) | _BV(3) | _BV(2);
  PORTC =  _BV(6);
  PORTD =  _BV(7) | _BV(6) | _BV(4) | _BV(2);
  PORTE =  _BV(6);
  PORTF =  _BV(7) | _BV(6) | _BV(5) | _BV(4) | _BV(1) | _BV(0);
// set output HIGH to all ROWs. BGLED off and CapsLED off.
  PORTB = PORTB | _BV(0) | _BV(1);
  PORTC = PORTC | _BV(7);
  PORTD = PORTD | _BV(0) | _BV(1) | _BV(3) | _BV(5);
  PORTE = PORTE | _BV(2);

  memset(key_pressed, 0, sizeof(key_pressed));
  memset(last_key_pressed, 0, sizeof(last_key_pressed));
  sei();

  Serial.begin(115200);
  while(!Serial)
    ;           // wait until CDC ready.
  Serial.println("scan_xd87 start.");    
  rgb_on(VAL_DIMMED);
  BGLED_ON();
  CAPSLED_OFF();
}

// すべての列を読んでビットマップに展開。
static void read_cols(uint8_t row) {
    key_pressed[row++] = ~( ((PINB << 2) & 0x80) | ((PINF >> 1) & 0x78) | ((PINF << 1) & 6) | ((PINE >> 6) & 1));
    key_pressed[row++] = ~( ((PINB << 4) & 0xc0) | ((PINB << 1) & 0x20) | 
          ((PIND >> 3) & 0x18) | ((PIND >> 2) & 4) | ((PINC >> 5) & 2) | ((PINB >> 6) & 1));
    key_pressed[row] = ~((PIND >> 2) & 1) & 1;
}

#define READ_ROW(port, pin, r)  { \
  port &= ~_BV(pin); \
  delayMicroseconds(5); \
  read_cols(r); \
  port |= _BV(pin); \
  delayMicroseconds(5); }

void scan() {
//  PORTD &= ~_BV(0); //  時間測定用トリガ
  READ_ROW(PORTD, 1, 0);
  READ_ROW(PORTB, 0, 3);
  READ_ROW(PORTB, 1, 6);
  READ_ROW(PORTC, 7, 9);
  READ_ROW(PORTD, 3, 12);
  READ_ROW(PORTD, 5, 15);
//  PORTD |= _BV(0);
}

unsigned long rgb_bright = 0;
int led_count = 0;
void loop() {
  scan();
  uint8_t code = 0;
  uint8_t m = 0;
  for(uint8_t i = 0; i < sizeof(key_pressed); i++) {
    uint8_t pressed = key_pressed[i];
    uint8_t delta = 8; 
    if (m == 2) {
      m = 0;
      delta = 1;
    } else
      m++;
    uint8_t change = pressed ^ last_key_pressed[i];
    if (change) {
      uint8_t mask = 1;
      for(uint8_t k = 0; k < delta; k++, mask <<= 1) {
        if (change & mask) {
          uint8_t c = code + k + 1;
          char tmp[10];
          sprintf(tmp, "%d , %d ", c, (pressed & mask) != 0);
          Serial.print(tmp);
          // rgb leds.
          rgb_on(VAL_BRIGHT);
          rgb_bright = millis();
        }
      }
      last_key_pressed[i] = pressed;
      Serial.println();
    }
    code += delta;
  }
  if (rgb_bright != 0 && (millis() - rgb_bright) > 200) {
    rgb_bright = 0;
    rgb_on(VAL_DIMMED);
  }
  delay(10);
  led_count++;
  if (led_count == 40) {
    BGLED_OFF();
    CAPSLED_ON();
  } else if (led_count == 120) {
    BGLED_ON();
    CAPSLED_OFF();
    led_count = 0;
  }
}
